#include <stdarg.h>
#include "SigsNetworkMessage.h"

#if WIN32
	#include <winsock2.h>
#else
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <unistd.h>
	#define closesocket(socket) close(socket)
#endif



struct BufferDescriptor
{
	int bufferSize;
	char *bufferData;
};


// Packs all the fields of a message into a buffer.
// The packing and the number of arguments used is determined based
// on the meta data associated with the given message type.
// Returns size of the message in bytes
// Returns 0 if an error occured (all valid messages are at least 12 bytes long)
int FormatMessage(void *messageBuffer, int maxSize, int messageType, ...)
{
	int i;
	int maxPtr = (int)messageBuffer + maxSize;
	char *curDataPtr = (char*)messageBuffer + sizeof(MessageHeader);
	va_list curArg;
	BufferDescriptor *tempBuffer;
	char *tempData;
	int tempLen;

	// Make sure there is at least enough buffer space for the header
	if (maxSize < sizeof(MessageHeader))
	{
		cerr << "Error: Server tried to create a message with insufficient buffer space for message header." << endl;
		return 0;	// No message constructed (not enough buffer space)
	}


	// Initialize message header fields
	((MessageHeader*)messageBuffer)->length = 0;
	((MessageHeader*)messageBuffer)->messageType = htonl(messageType);
	((MessageHeader*)messageBuffer)->flags = 0x01;


	// Lookup the message format
	const MessageDataFormat *format = FindMessageDataFormat(messageType);
	// Check if the format wasn't found
	if (format == NULL)
	{
		cerr << "Error: Server tried to create a message with an unknown format." << endl;
		return 0;	// No message constructed (unknown format)
	}

	// Process all the function arguments/message fields
	va_start(curArg, messageType);
	for (i = 0; i < format->numDataFields; i++)
	{
		switch(format->dataTypes[i])
		{
			case Word:		// Short (16 bit, no endian conversion required)
				// Check if buffer has space for the data
				if (maxPtr - (int)curDataPtr < sizeof(short))
					return 0;

				*(short*)curDataPtr = va_arg(curArg, short);
				curDataPtr += sizeof(short);

				break;
			case Int:		// Big Endian int (32 bit)
				// Check if buffer has space for the data
				if (maxPtr - (int)curDataPtr < sizeof(int))
					return 0;

				*(int*)curDataPtr = ntohl(va_arg(curArg, int));
				curDataPtr += sizeof(int);

				break;
			case Float:
				// Check if buffer has space for the data
				if (maxPtr - (int)curDataPtr < sizeof(float))
					return 0;

				*(float*)curDataPtr = va_arg(curArg, float);
				curDataPtr += sizeof(float);

				break;
			case Double:
				// Check if buffer has space for the data
				if (maxPtr - (int)curDataPtr < sizeof(double))
					return 0;

				*(double*)curDataPtr = va_arg(curArg, double);
				curDataPtr += sizeof(double);

				break;
			case Short:		// Big Endian short (16 bit)
				// Check if buffer has space for the data
				if (maxPtr - (int)curDataPtr < sizeof(short))
					return 0;

				*(short*)curDataPtr = ntohs(va_arg(curArg, short));
				curDataPtr += sizeof(short);

				break;
			case Dword:		// Int (32 bit, no endian conversion required)
				// Check if buffer has space for the data
				if (maxPtr - (int)curDataPtr < sizeof(int))
					return 0;

				*(int*)curDataPtr = va_arg(curArg, int);
				curDataPtr += sizeof(int);

				break;
			case Char:		// Single byte (char)
				// Check if buffer has space for the data
				if (maxPtr - (int)curDataPtr < sizeof(char))
					return 0;

				*(char*)curDataPtr = va_arg(curArg, char);
				curDataPtr += sizeof(char);

				break;
			case String:	// NULL terminated ASCII string
				tempData = va_arg(curArg, char*);	// Get the string (pointer) parameter
				tempLen = strlen(tempData);			// Calculate string length
				
				// Check if buffer has space left for the string
				if (maxPtr - (int)curDataPtr <= tempLen)
					return 0;
				
				// Copy the string to the buffer
				strcpy(curDataPtr, tempData);
				curDataPtr += tempLen + 1;

				break;
			case Buffer:	// Length prefixed buffer
				tempBuffer = va_arg(curArg, BufferDescriptor*);	// Get BufferDescriptor (pointer) parameter

				// Check if buffer has space for the data
				if (maxPtr - (int)curDataPtr < (int)(tempBuffer->bufferSize + sizeof(int)))
					return 0;

				// Copy the buffer size to the message
				*(int*)curDataPtr = tempBuffer->bufferSize;
				curDataPtr += sizeof(int);
				// Copy the buffer contents to the message
				memcpy(curDataPtr, tempBuffer->bufferData, tempBuffer->bufferSize);
				curDataPtr += tempBuffer->bufferSize;

				break;
			default:		// Unexpected data format error
				break;
		}
	}

	// Calculate the length of the message buffer
	tempLen = (int)curDataPtr - (int)messageBuffer;
	// Set the header length field
	((MessageHeader*)messageBuffer)->length = htonl(tempLen);

	// Return the length of the created message
	return tempLen;
}


void DumpHumanReadableMessage(ostream &outStream, void *messageBuffer)
{
	int i, j;
	char *curDataPtr = (char*)messageBuffer + sizeof(MessageHeader);
	MessageHeader &header = *(MessageHeader*)messageBuffer;
	int maxPtr = (int)messageBuffer + ntohl(header.length);
	int tempLength;


	// Dispaly the message type
	outStream << "MessageType: " << ntohl(header.messageType);

	// Lookup the message format
	const MessageDataFormat *format = FindMessageDataFormat(ntohl(header.messageType));
	// Check if the format wasn't found
	if (format == NULL)
	{
		// Error, unknown format
		outStream << endl << "Unknown message format" << endl;

		// Output a binary dump of the message
		for (j = 12; j < header.length; j++)
		{
			outStream << *curDataPtr << " ";
			curDataPtr++;
		}
		outStream << endl;
		
		// Can't continue (unknown format)
		return;
	}
	outStream << " (Number of data fields: " << format->numDataFields << ")" << endl;


	// Display all the data fields
	for (i = 0; i < format->numDataFields; i++)
	{
		switch(format->dataTypes[i])
		{
		case Word:		// Short (16 bit, no endian conversion required)
			// Check if buffer has space for the data
			if (maxPtr - (int)curDataPtr < sizeof(short))
			{
				outStream << "Warning: Message size exceeded." << endl;
				return;
			}

			outStream << *(short*)curDataPtr << endl;
			curDataPtr += sizeof(short);

			break;
		case Int:		// Big Endian int (32 bit)
			// Check if buffer has space for the data
			if (maxPtr - (int)curDataPtr < sizeof(int))
			{
				outStream << "Warning: Message size exceeded." << endl;
				return;
			}

			outStream << ntohl(*(int*)curDataPtr) << endl;
			curDataPtr += sizeof(int);

			break;
		case Float:
			// Check if buffer has space for the data
			if (maxPtr - (int)curDataPtr < sizeof(float))
			{
				outStream << "Warning: Message size exceeded." << endl;
				return;
			}

			outStream << *(float*)curDataPtr << endl;
			curDataPtr += sizeof(float);

			break;
		case Double:
			// Check if buffer has space for the data
			if (maxPtr - (int)curDataPtr < sizeof(double))
			{
				outStream << "Warning: Message size exceeded." << endl;
				return;
			}

			outStream << *(double*)curDataPtr << endl;
			curDataPtr += sizeof(double);

			break;
		case Short:		// Big Endian short (16 bit)
			// Check if buffer has space for the data
			if (maxPtr - (int)curDataPtr < sizeof(short))
			{
				outStream << "Warning: Message size exceeded." << endl;
				return;
			}

			outStream << ntohs(*(short*)curDataPtr) << endl;
			curDataPtr += sizeof(short);

			break;
		case Dword:		// Int (32 bit, no endian conversion required)
			// Check if buffer has space for the data
			if (maxPtr - (int)curDataPtr < sizeof(int))
			{
				outStream << "Warning: Message size exceeded." << endl;
				return;
			}

			outStream << *(int*)curDataPtr << endl;
			curDataPtr += sizeof(int);

			break;
		case Char:		// Single byte (char)
			// Check if buffer has space for the data
			if (maxPtr - (int)curDataPtr < sizeof(char))
			{
				outStream << "Warning: Message size exceeded." << endl;
				return;
			}

			outStream << (int)(*curDataPtr) << endl;
			curDataPtr += sizeof(char);

			break;
		case String:	// NULL terminated ASCII string
			// Find the length of the string
			tempLength = strlen(curDataPtr);

			// Check if buffer has space for the data
			if (maxPtr - (int)curDataPtr <= tempLength)
			{
				outStream << "Warning: Message size exceeded." << endl;
				return;
			}

			outStream << curDataPtr << endl;
			curDataPtr += tempLength + 1;

			break;
		case Buffer:	// Length prefixed buffer
			// Check if the buffer has space for the length prefix
			if (maxPtr - (int)curDataPtr < sizeof(int))
			{
				outStream << "Warning: Message size exceeded." << endl;
				return;
			}
			tempLength = *(int*)curDataPtr;
			outStream << "Buffer length: " << tempLength << endl;
			curDataPtr += sizeof(int);

			// Check if the (message) buffer has space for all of the buffer data
			if (maxPtr - (int)curDataPtr < tempLength)
			{
				outStream << "Warning: Message size exceeded." << endl;
				return;
			}

			// Output a binary dump of the message
			for (j = 12; j < header.length; j++)
			{
				outStream << *curDataPtr << " ";
				curDataPtr++;
			}
			outStream << endl;

			break;
		default:		// Unexpected data format error
			outStream << "Warning: Unexpected data type in message format." << endl;

			break;
		}
	}
}



int SendMessage(int sock, void *messageBuffer)
{
	MessageHeader &header = *(MessageHeader*)messageBuffer;
	int errorCode;

	// Display a debug message
	//cout << "Responding with message:" << endl;
	//DumpHumanReadableMessage(cout, messageBuffer);

	// Send the message length over the socket
	errorCode = send(sock, (char*)&header.length, 4, 0);
	// Send the message over the socket
	errorCode = send(sock, (char*)messageBuffer, ntohl(header.length), 0);

	return errorCode;
}


